"""One-component GEMC through RaspaBaseWorkChain"""

import sys

import click
from aiida.common import NotExistent
from aiida.engine import run_get_node
from aiida.orm import Code, Dict

from aiida_raspa.workchains import RaspaBaseWorkChain


def example_base_workchain_gemc(raspa_code):
    """Run the base workchain for GEMC calculation."""

    # pylint: disable=no-member

    print("Testing RASPA methane GEMC through RaspaBaseWorkChain ...")

    # parameters
    parameters = Dict(
        dict={
            "GeneralSettings": {
                "SimulationType": "MonteCarlo",
                "NumberOfCycles": 50,
                "NumberOfInitializationCycles": 50,
                "PrintEvery": 10,
                "Forcefield": "GenericMOFs",
                "CutOff": 12.0,
                "GibbsVolumeChangeProbability": 0.1,
            },
            "System": {
                "box_one": {
                    "type": "Box",
                    "BoxLengths": "30 30 30",
                    "BoxAngles": "90 90 90",
                    "ExternalTemperature": 300.0,
                },
                "box_two": {
                    "type": "Box",
                    "BoxLengths": "30 30 30",
                    "BoxAngles": "90 90 90",
                    "ExternalTemperature": 300.0,
                },
            },
            "Component": {
                "methane": {
                    "MoleculeDefinition": "TraPPE",
                    "TranslationProbability": 1.0,
                    "ReinsertionProbability": 1.0,
                    "GibbsSwapProbability": 1.0,
                    "CreateNumberOfMolecules": {
                        "box_one": 150,
                        "box_two": 150,
                    },
                },
            },
        }
    )

    # Constructing builder
    builder = RaspaBaseWorkChain.get_builder()

    # Specifying the code
    builder.raspa.code = raspa_code

    # Specifying the input parameters
    builder.raspa.parameters = parameters

    # Add handlers that could handle physics-related problems.
    builder.handler_overrides = Dict(
        dict={
            "check_gemc_box": True,
            "check_gemc_convergence": True,
        }
    )  # Enable gemc handlers disabled by default.

    # Specifying the scheduler options
    builder.raspa.metadata.options = {
        "resources": {
            "num_machines": 1,
            "num_mpiprocs_per_machine": 1,
        },
        "max_wallclock_seconds": 1 * 30 * 60,  # 30 min
        "withmpi": False,
    }

    _, node = run_get_node(builder)
    assert node.exit_status == 0


@click.command("cli")
@click.argument("codelabel")
def cli(codelabel):
    """Click interface"""
    try:
        code = Code.get_from_string(codelabel)
    except NotExistent:
        print(f"The code '{codelabel}' does not exist")
        sys.exit(1)
    example_base_workchain_gemc(code)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter

# EOF



"""Run RASPA calculation using Local force field"""
import os
import sys

import click
from aiida.common import NotExistent
from aiida.engine import run, run_get_pk
from aiida.orm import CifData, Code, Dict, SinglefileData


def example_ff_files(raspa_code, submit=True):
    """Prepare and submit RASPA calculation with components mixture."""

    # parameters
    parameters = Dict(
        dict={
            "GeneralSettings": {
                "SimulationType": "MonteCarlo",
                "NumberOfInitializationCycles": 50,
                "NumberOfCycles": 50,
                "PrintEvery": 10,
                "Forcefield": "Local",
                "ChargeMethod": "Ewald",
                "EwaldPrecision": 1e-6,
                "CutOff": 12.0,
            },
            "System": {
                "irmof_1": {
                    "type": "Framework",
                    "UnitCells": "1 1 1",
                    "HeliumVoidFraction": 0.149,
                    "ExternalTemperature": 300.0,
                    "ExternalPressure": 1e5,
                }
            },
            "Component": {
                "CO2": {
                    "MoleculeDefinition": "Local",
                    "MolFraction": 0.30,
                    "TranslationProbability": 1.0,
                    "RotationProbability": 1.0,
                    "ReinsertionProbability": 1.0,
                    "SwapProbability": 1.0,
                    "CreateNumberOfMolecules": 0,
                },
                "N2": {
                    "MoleculeDefinition": "Local",
                    "MolFraction": 0.70,
                    "TranslationProbability": 1.0,
                    "RotationProbability": 1.0,
                    "ReinsertionProbability": 1.0,
                    "SwapProbability": 1.0,
                    "CreateNumberOfMolecules": 0,
                },
            },
        }
    )

    # Contructing builder
    pwd = os.path.dirname(os.path.realpath(__file__))
    builder = raspa_code.get_builder()
    builder.framework = {
        "irmof_1": CifData(file=os.path.join(pwd, "..", "files", "IRMOF-1_eqeq.cif")),
    }
    # Note: Here the SinglefileData in the dict are stored otherwise the dry_run crashes.
    #       However, this is not needed for real calculations (e.g., using --submit), since the work chains stores them.
    builder.file = {
        "file_1": SinglefileData(file=os.path.join(pwd, "..", "files", "force_field_mixing_rules.def")).store(),
        "file_2": SinglefileData(file=os.path.join(pwd, "..", "files", "pseudo_atoms.def")).store(),
        "file_3": SinglefileData(file=os.path.join(pwd, "..", "files", "CO2.def")).store(),
        "file_4": SinglefileData(file=os.path.join(pwd, "..", "files", "N2.def")).store(),
    }
    builder.parameters = parameters
    builder.metadata.options = {
        "resources": {
            "num_machines": 1,
            "num_mpiprocs_per_machine": 1,
        },
        "max_wallclock_seconds": 1 * 30 * 60,  # 30 min
        "withmpi": False,
    }
    builder.metadata.dry_run = False
    builder.metadata.store_provenance = True

    if submit:
        print("Testing RASPA CO2/N2 adsorption in IRMOF-1, using Local force field ...")
        res, pk = run_get_pk(builder)
        print("calculation pk: ", pk)
        # pylint: disable=consider-using-f-string
        print(
            "CO2/N2 uptake ({:s}): {:.2f}/{:.2f} ".format(
                res["output_parameters"]["irmof_1"]["components"]["N2"]["loading_absolute_unit"],
                res["output_parameters"]["irmof_1"]["components"]["CO2"]["loading_absolute_average"],
                res["output_parameters"]["irmof_1"]["components"]["N2"]["loading_absolute_average"],
            )
        )
        print("OK, calculation has completed successfully")
    else:
        print("Generating test input ...")
        builder.metadata.dry_run = True
        builder.metadata.store_provenance = False
        run(builder)
        print("Submission test successful")
        print("In order to actually submit, add '--submit'")
    print("-----")


@click.command("cli")
@click.argument("codelabel")
@click.option("--submit", is_flag=True, help="Actually submit calculation")
def cli(codelabel, submit):
    """Click interface"""
    try:
        code = Code.get_from_string(codelabel)
    except NotExistent:
        print(f"The code '{codelabel}' does not exist")
        sys.exit(1)
    example_ff_files(code, submit)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter

# EOF



"""Basic raspa output parser."""
import re
from math import isinf, isnan

float_base = float  # pylint: disable=invalid-name


def float(number):  # pylint: disable=redefined-builtin
    number = float_base(number)
    return number if not any((isnan(number), isinf(number))) else None


KELVIN_TO_KJ_PER_MOL = float(8.314464919 / 1000.0)  # exactly the same as Raspa

# manage block of the first type
# --------------------------------------------------------------------------------------------
BLOCK_1_LIST = [
    # ("Average temperature:", "temperature", (1, 2, 4), 0), # misleading property!
    # ("Average Pressure:", "pressure", (1, 2, 4), 0), # misleading property!
    ("Average Volume:", "cell_volume", (1, 2, 4), 0),
    ("Average Density:", "adsorbate_density", (1, 2, 4), 0),
    # ("Average Heat Capacity", "framework_heat_capacity", (1, 2, 4), 0), # misleading property!
    ("Enthalpy of adsorption:", "enthalpy_of_adsorption", (1, 4, 3), 4),
    ("Tail-correction energy:", "tail_correction_energy", (1, 2, 4), 0)
    # ("Total energy:$", "total_energy", (1, 2, 4), 0), # not important property!
]

# block of box properties.
BOX_PROP_LIST = [
    ("Average Box-lengths:", "box"),
]


# pylint: disable=too-many-arguments
def parse_block1(flines, result_dict, prop, value=1, unit=2, dev=4):
    """Parse block.

    Parses blocks that look as follows::

        Average Volume:
        =================
            Block[ 0]        12025.61229 [A^3]
            Block[ 1]        12025.61229 [A^3]
            Block[ 2]        12025.61229 [A^3]
            Block[ 3]        12025.61229 [A^3]
            Block[ 4]        12025.61229 [A^3]
            ------------------------------------------------------------------------------
            Average          12025.61229 [A^3] +/-            0.00000 [A^3]

    """
    for line in flines:
        if "Average" in line:
            result_dict[prop + "_average"] = float(line.split()[value])
            result_dict[prop + "_unit"] = re.sub(r"[{}()\[\]]", "", line.split()[unit])
            result_dict[prop + "_dev"] = float(line.split()[dev])
            break


# manage energy reading
# --------------------------------------------------------------------------------------------
ENERGY_CURRENT_LIST = [
    ("Host/Adsorbate energy:", "host/ads", "tot"),
    ("Host/Adsorbate VDW energy:", "host/ads", "vdw"),
    ("Host/Adsorbate Coulomb energy:", "host/ads", "coulomb"),
    ("Adsorbate/Adsorbate energy:", "ads/ads", "tot"),
    ("Adsorbate/Adsorbate VDW energy:", "ads/ads", "vdw"),
    ("Adsorbate/Adsorbate Coulomb energy:", "ads/ads", "coulomb"),
]

ENERGY_AVERAGE_LIST = [
    ("Average Adsorbate-Adsorbate energy:", "ads/ads"),
    ("Average Host-Adsorbate energy:", "host/ads"),
]


def parse_block_energy(flines, res_dict, prop):
    """Parse energy block.

    Parse block that looks as follows::

        Average Adsorbate-Adsorbate energy:
        ===================================
            Block[ 0] -443.23204         Van der Waals: -443.23204         Coulomb: 0.00000            [K]
            Block[ 1] -588.20205         Van der Waals: -588.20205         Coulomb: 0.00000            [K]
            Block[ 2] -538.43355         Van der Waals: -538.43355         Coulomb: 0.00000            [K]
            Block[ 3] -530.00960         Van der Waals: -530.00960         Coulomb: 0.00000            [K]
            Block[ 4] -484.15106         Van der Waals: -484.15106         Coulomb: 0.00000            [K]
            ------------------------------------------------------------------------------
            Average   -516.80566         Van der Waals: -516.805659        Coulomb: 0.00000            [K]
                  +/- 98.86943                      +/- 98.869430               +/- 0.00000            [K]
    """
    for line in flines:
        if "Average" in line:
            res_dict[f"energy_{prop}_tot_average"] = float(line.split()[1]) * KELVIN_TO_KJ_PER_MOL
            res_dict[f"energy_{prop}_vdw_average"] = float(line.split()[5]) * KELVIN_TO_KJ_PER_MOL
            res_dict[f"energy_{prop}_coulomb_average"] = float(line.split()[7]) * KELVIN_TO_KJ_PER_MOL
        if "+/-" in line:
            res_dict[f"energy_{prop}_tot_dev"] = float(line.split()[1]) * KELVIN_TO_KJ_PER_MOL
            res_dict[f"energy_{prop}_vdw_dev"] = float(line.split()[3]) * KELVIN_TO_KJ_PER_MOL
            res_dict[f"energy_{prop}_coulomb_dev"] = float(line.split()[5]) * KELVIN_TO_KJ_PER_MOL
            return


# manage lines with components
# --------------------------------------------------------------------------------------------
LINES_WITH_COMPONENT_LIST = [
    (" Average Widom Rosenbluth-weight:", "widom_rosenbluth_factor"),
    (" Average chemical potential: ", "chemical_potential"),
    (" Average Henry coefficient: ", "henry_coefficient"),
    (" Average  <U_gh>_1-<U_h>_0:", "adsorption_energy_widom"),
]


def parse_lines_with_component(res_components, components, line, prop):
    """Parse lines that contain components"""
    # self.logger.info("analysing line: {}".format(line))
    for i, component in enumerate(components):
        if "[" + component + "]" in line:
            words = line.split()
            res_components[i][prop + "_unit"] = re.sub(r"[{}()\[\]]", "", words[-1])
            res_components[i][prop + "_dev"] = float(words[-2])
            res_components[i][prop + "_average"] = float(words[-4])


# pylint: disable=too-many-locals, too-many-arguments, too-many-statements, too-many-branches
def parse_base_output(output_contents, system_name, ncomponents):
    """Parse RASPA output file: it is divided in different parts, whose start/end is carefully documented."""

    warnings = []
    res_per_component = []
    for i in range(ncomponents):
        res_per_component.append({})
    result_dict = {"exceeded_walltime": False}

    output_contents = iter(output_contents.split("\n"))

    # 1st parsing part: input settings
    # --------------------------------
    # from: start of file
    # to: "Current (initial full energy) Energy Status"
    icomponent = 0
    component_names = []
    res_cmp = res_per_component[0]

    for line in output_contents:
        if "Component" in line and "molecule)" in line:
            component_names.append(line.split()[2][1:-1])
            if "(Adsorbate" in line:
                res_cmp["molecule_type"] = "adsorbate"
            elif "(Cation" in line:
                res_cmp["molecule_type"] = "cation"

        # Consider to change it with parse_line()
        if "Conversion factor molecules/unit cell -> mol/kg:" in line:
            res_cmp["conversion_factor_molec_uc_to_mol_kg"] = float(line.split()[6])
            res_cmp["conversion_factor_molec_uc_to_mol_kg_unit"] = "(mol/kg)/(molec/uc)"
        # this line was corrected in Raspa's commit c1ad4de (Nov19), since "gr/gr" should read "mg/g"
        if (
            "Conversion factor molecules/unit cell -> gr/gr:" in line
            or "Conversion factor molecules/unit cell -> mg/g:" in line
        ):
            res_cmp["conversion_factor_molec_uc_to_mg_g"] = float(line.split()[6])
            res_cmp["conversion_factor_molec_uc_to_mg_g_unit"] = "(mg/g)/(molec/uc)"
        if "Conversion factor molecules/unit cell -> cm^3 STP/gr:" in line:
            res_cmp["conversion_factor_molec_uc_to_cm3stp_gr"] = float(line.split()[7])
            res_cmp["conversion_factor_molec_uc_to_cm3stp_gr_unit"] = "(cm^3_STP/gr)/(molec/uc)"
        if "Conversion factor molecules/unit cell -> cm^3 STP/cm^3:" in line:
            res_cmp["conversion_factor_molec_uc_to_cm3stp_cm3"] = float(line.split()[7])
            res_cmp["conversion_factor_molec_uc_to_cm3stp_cm3_unit"] = "(cm^3_STP/cm^3)/(molec/uc)"
        if "MolFraction:" in line:
            res_cmp["mol_fraction"] = float(line.split()[1])
            res_cmp["mol_fraction_unit"] = "-"
        if "Partial pressure:" in line:
            res_cmp["partial_pressure"] = float(line.split()[2])
            res_cmp["partial_pressure_unit"] = "Pa"
        if "Partial fugacity:" in line:
            res_cmp["partial_fugacity"] = float(line.split()[2])
            res_cmp["partial_fugacity_unit"] = "Pa"
            icomponent += 1
            if icomponent < ncomponents:
                res_cmp = res_per_component[icomponent]
        if "Framework Density" in line:
            result_dict["framework_density"] = line.split()[2]
            result_dict["framework_density_unit"] = re.sub(r"[{}()\[\]]", "", line.split()[3])
        if "Current (initial full energy) Energy Status" in line:
            break

    # 2nd parsing part: initial and final configurations
    # --------------------------------------------------
    # from: "Current (initial full energy) Energy Status"
    # to: "Average properties of the system"

    reading = "initial"
    result_dict["energy_unit"] = "kJ/mol"

    for line in output_contents:
        # Understand if it is the initial or final "Current Energy Status" section
        if "Current (full final energy) Energy Status" in line:
            reading = "final"

        # Read the entries of "Current Energy Status" section
        if reading:
            for parse in ENERGY_CURRENT_LIST:
                if parse[0] in line:
                    result_dict[f"energy_{parse[1]}_{parse[2]}_{reading}"] = (
                        float(line.split()[-1]) * KELVIN_TO_KJ_PER_MOL
                    )
                    if parse[1] == "ads/ads" and parse[2] == "coulomb":
                        reading = None

        if "Average properties of the system" in line:
            break

    # 3rd parsing part: average system properties
    # --------------------------------------------------
    # from: "Average properties of the system"
    # to: "Number of molecules"

    for line in output_contents:
        # ("Enthalpy of adsorption:", "enthalpy_of_adsorption", (1, 4, 3), 4),
        for parse in BLOCK_1_LIST:
            if parse[0] in line:
                parse_block1(output_contents, result_dict, parse[1], *parse[2])
                # I assume here that properties per component are present furhter in the output file.
                # so I need to skip some lines:
                skip_nlines_after = parse[3]
                while skip_nlines_after > 0:
                    line = next(output_contents)
                    skip_nlines_after -= 1
                for i, cmpnt in enumerate(component_names):
                    # The order of properties per molecule is the same as the order of molecules in the
                    # input file. So if component name was not found in the next line, I break the loop
                    # immidiately as there is no reason to continue it
                    line = next(output_contents)
                    if cmpnt in line:
                        parse_block1(output_contents, res_per_component[i], parse[1], *parse[2])
                    else:
                        break
                    skip_nlines_after = parse[3]
                    while skip_nlines_after > 0:
                        line = next(output_contents)
                        skip_nlines_after -= 1

                continue  # no need to perform further checks, propperty has been found already
        for parse in ENERGY_AVERAGE_LIST:
            if parse[0] in line:
                parse_block_energy(output_contents, result_dict, prop=parse[1])
                continue  # no need to perform further checks, propperty has been found already
        for parse in BOX_PROP_LIST:
            if parse[0] in line:
                # parse three cell vectors
                parse_block1(output_contents, result_dict, prop="box_ax", value=2, unit=3, dev=5)
                parse_block1(output_contents, result_dict, prop="box_by", value=2, unit=3, dev=5)
                parse_block1(output_contents, result_dict, prop="box_cz", value=2, unit=3, dev=5)
                # parsee angles between the cell vectors
                parse_block1(output_contents, result_dict, prop="box_alpha", value=3, unit=4, dev=6)
                parse_block1(output_contents, result_dict, prop="box_beta", value=3, unit=4, dev=6)
                parse_block1(output_contents, result_dict, prop="box_gamma", value=3, unit=4, dev=6)

        if "Number of molecules:" in line:
            break

    # 4th parsing part: average molecule properties
    # --------------------------------------------------
    # from: "Number of molecules"
    # to: end of file

    icomponent = 0
    for line in output_contents:
        # Consider to change it with parse_line?
        if "Average loading absolute [molecules/unit cell]" in line:
            res_per_component[icomponent]["loading_absolute_average"] = float(line.split()[5])
            res_per_component[icomponent]["loading_absolute_dev"] = float(line.split()[7])
            res_per_component[icomponent]["loading_absolute_unit"] = "molecules/unit cell"
        elif "Average loading excess [molecules/unit cell]" in line:
            res_per_component[icomponent]["loading_excess_average"] = float(line.split()[5])
            res_per_component[icomponent]["loading_excess_dev"] = float(line.split()[7])
            res_per_component[icomponent]["loading_excess_unit"] = "molecules/unit cell"
            icomponent += 1
        if icomponent >= ncomponents:
            break

    for line in output_contents:
        for to_parse in LINES_WITH_COMPONENT_LIST:
            if to_parse[0] in line:
                parse_lines_with_component(res_per_component, component_names, line, to_parse[1])

    # Assigning to None all the quantities that are meaningless if not running a Widom insertion calculation
    for res_comp in res_per_component:
        for prop in ["henry_coefficient", "widom_rosenbluth_factor", "chemical_potential"]:
            if res_comp[f"{prop}_dev"] == 0.0:
                res_comp[f"{prop}_average"] = None
                res_comp[f"{prop}_dev"] = None

        # The section "Adsorption energy from Widom-insertion" is not showing in the output if no widom is performed
        if not "adsorption_energy_widom_average" in res_comp:
            res_comp["adsorption_energy_widom_unit"] = "kJ/mol"
            res_comp["adsorption_energy_widom_dev"] = None
            res_comp["adsorption_energy_widom_average"] = None

    return_dictionary = {"general": result_dict, "components": {}}

    for name, value in zip(component_names, res_per_component):
        return_dictionary["components"][name] = value

    # Parsing all the warning that are printed in the output file, avoiding redoundancy
    for line in output_contents:
        if "WARNING" in line:
            warning_touple = (system_name, line)
            if warning_touple not in warnings:
                warnings.append(warning_touple)

    return return_dictionary, warnings



"""Test Raspa output parser"""

import os
from pathlib import Path

from aiida_raspa.utils import parse_base_output

CWD = os.path.dirname(os.path.realpath(__file__))


def test_parse_output_one_cmpnt():
    """Testing output parser on an output file with one component"""

    with Path(CWD, "outputs/one_component.out").open("r", encoding="utf-8") as handle:
        parsed_parameters = parse_base_output(handle.read(), system_name="system1", ncomponents=1)[0]

    general = {
        "exceeded_walltime": False,
        "framework_density": "739.995779685958",
        "framework_density_unit": "kg/m^3",
        "energy_unit": "kJ/mol",
        "energy_host/ads_tot_initial": 0.0,
        "energy_host/ads_vdw_initial": 0.0,
        "energy_host/ads_coulomb_initial": 0.0,
        "energy_host/ads_tot_final": -171.717724280208,
        "energy_host/ads_vdw_final": -171.717724280208,
        "energy_host/ads_coulomb_final": 0.0,
        "cell_volume_average": 12025.61229,
        "cell_volume_unit": "A^3",
        "cell_volume_dev": 0.0,
        "box_ax_average": 34.1995,
        "box_ax_unit": "A^3",
        "box_ax_dev": 0.0,
        "box_by_average": 22.6557,
        "box_by_unit": "A^3",
        "box_by_dev": 0.0,
        "box_cz_average": 15.52065,
        "box_cz_unit": "A^3",
        "box_cz_dev": 0.0,
        "box_alpha_average": 90.0,
        "box_alpha_unit": "degrees",
        "box_alpha_dev": 0.0,
        "box_beta_average": 52.8626,
        "box_beta_unit": "degrees",
        "box_beta_dev": 0.0,
        "box_gamma_average": 90.0,
        "box_gamma_unit": "degrees",
        "box_gamma_dev": 0.0,
        "adsorbate_density_average": 26.62672,
        "adsorbate_density_unit": "kg/m^3",
        "adsorbate_density_dev": 5.35266,
        "enthalpy_of_adsorption_average": -1928.58214,
        "enthalpy_of_adsorption_unit": "K",
        "enthalpy_of_adsorption_dev": 204.730157,
        "energy_ads/ads_tot_average": -4.5501817208796655,
        "energy_ads/ads_vdw_average": -4.550181687621805,
        "energy_ads/ads_coulomb_average": 0.0,
        "energy_ads/ads_tot_dev": 2.0057559085076093,
        "energy_ads/ads_vdw_dev": 2.0057559334510042,
        "energy_ads/ads_coulomb_dev": 0.0,
        "energy_host/ads_tot_average": -174.60996722014394,
        "energy_host/ads_vdw_average": -174.60996720351505,
        "energy_host/ads_coulomb_average": 0.0,
        "energy_host/ads_tot_dev": 34.12887304355653,
        "energy_host/ads_vdw_dev": 34.12887304355653,
        "energy_host/ads_coulomb_dev": 0.0,
        "tail_correction_energy_average": 0.0,
        "tail_correction_energy_unit": "K",
        "tail_correction_energy_dev": 0.0,
    }

    methane = {
        "mol_fraction": 1.0,
        "mol_fraction_unit": "-",
        "conversion_factor_molec_uc_to_mol_kg": 0.1866003989,
        "conversion_factor_molec_uc_to_mol_kg_unit": "(mol/kg)/(molec/uc)",
        "conversion_factor_molec_uc_to_mg_g": 2.9935294361,
        "conversion_factor_molec_uc_to_mg_g_unit": "(mg/g)/(molec/uc)",
        "conversion_factor_molec_uc_to_cm3stp_gr": 4.1824568165,
        "conversion_factor_molec_uc_to_cm3stp_gr_unit": "(cm^3_STP/gr)/(molec/uc)",
        "conversion_factor_molec_uc_to_cm3stp_cm3": 3.095000393,
        "conversion_factor_molec_uc_to_cm3stp_cm3_unit": "(cm^3_STP/cm^3)/(molec/uc)",
        "partial_pressure": 500000.0,
        "partial_pressure_unit": "Pa",
        "partial_fugacity": 494612.1383597442,
        "partial_fugacity_unit": "Pa",
        "adsorbate_density_average": 26.62672,
        "adsorbate_density_unit": "kg/m^3",
        "adsorbate_density_dev": 5.35266,
        "loading_absolute_average": 12.02,
        "loading_absolute_dev": 2.4163298616,
        "loading_absolute_unit": "molecules/unit cell",
        "loading_excess_average": 11.02,
        "loading_excess_dev": 2.4163298616,
        "loading_excess_unit": "molecules/unit cell",
        "widom_rosenbluth_factor_unit": "-",
        "widom_rosenbluth_factor_dev": None,
        "widom_rosenbluth_factor_average": None,
        "chemical_potential_unit": "K",
        "chemical_potential_dev": None,
        "chemical_potential_average": None,
        "henry_coefficient_unit": "mol/kg/Pa",
        "henry_coefficient_dev": None,
        "henry_coefficient_average": None,
        "adsorption_energy_widom_unit": "kJ/mol",
        "adsorption_energy_widom_dev": None,
        "adsorption_energy_widom_average": None,
    }

    for key, value in general.items():
        assert value == parsed_parameters["general"][key]

    for key, value in methane.items():
        assert value == parsed_parameters["components"]["methane"][key]


def test_parse_output_two_cmpnt():
    """Testing output parser on an output file with two components"""

    with Path(CWD, "outputs/two_components.out").open("r", encoding="utf-8") as handle:
        parsed_parameters = parse_base_output(handle.read(), system_name="system1", ncomponents=2)[0]

    general = {
        "exceeded_walltime": False,
        "framework_density": "0.000000000000",
        "framework_density_unit": "kg/m^3",
        "energy_unit": "kJ/mol",
        "energy_host/ads_tot_initial": 0.0,
        "energy_host/ads_vdw_initial": 0.0,
        "energy_host/ads_coulomb_initial": 0.0,
        "energy_host/ads_tot_final": 0.0,
        "energy_host/ads_vdw_final": 0.0,
        "energy_host/ads_coulomb_final": 0.0,
        "cell_volume_average": 15625.0,
        "cell_volume_unit": "A^3",
        "cell_volume_dev": 0.0,
        "box_ax_average": 25.0,
        "box_ax_unit": "A^3",
        "box_ax_dev": 0.0,
        "box_by_average": 25.0,
        "box_by_unit": "A^3",
        "box_by_dev": 0.0,
        "box_cz_average": 25.0,
        "box_cz_unit": "A^3",
        "box_cz_dev": 0.0,
        "box_alpha_average": 90.0,
        "box_alpha_unit": "degrees",
        "box_alpha_dev": 0.0,
        "box_beta_average": 90.0,
        "box_beta_unit": "degrees",
        "box_beta_dev": 0.0,
        "box_gamma_average": 90.0,
        "box_gamma_unit": "degrees",
        "box_gamma_dev": 0.0,
        "adsorbate_density_average": 4.40397,
        "adsorbate_density_unit": "kg/m^3",
        "adsorbate_density_dev": 0.74921,
        "enthalpy_of_adsorption_average": 137.68727,
        "enthalpy_of_adsorption_unit": "K",
        "enthalpy_of_adsorption_dev": 86.158033,
        "energy_ads/ads_tot_average": -0.15252104334203112,
        "energy_ads/ads_vdw_average": -0.15252104334203112,
        "energy_ads/ads_coulomb_average": 0.0,
        "energy_ads/ads_tot_dev": 0.16141328042825245,
        "energy_ads/ads_vdw_dev": 0.16141332200057704,
        "energy_ads/ads_coulomb_dev": 0.0,
        "energy_host/ads_tot_average": 0.0,
        "energy_host/ads_vdw_average": 0.0,
        "energy_host/ads_coulomb_average": 0.0,
        "energy_host/ads_tot_dev": 0.0,
        "energy_host/ads_vdw_dev": 0.0,
        "energy_host/ads_coulomb_dev": 0.0,
        "tail_correction_energy_average": 0.0,
        "tail_correction_energy_unit": "K",
        "tail_correction_energy_dev": 0.0,
    }

    butane = {
        "mol_fraction": 0.5,
        "mol_fraction_unit": "-",
        "conversion_factor_molec_uc_to_mol_kg": None,
        "conversion_factor_molec_uc_to_mol_kg_unit": "(mol/kg)/(molec/uc)",
        "conversion_factor_molec_uc_to_mg_g": None,
        "conversion_factor_molec_uc_to_mg_g_unit": "(mg/g)/(molec/uc)",
        "conversion_factor_molec_uc_to_cm3stp_gr": None,
        "conversion_factor_molec_uc_to_cm3stp_gr_unit": "(cm^3_STP/gr)/(molec/uc)",
        "conversion_factor_molec_uc_to_cm3stp_cm3": 2.3820335839,
        "conversion_factor_molec_uc_to_cm3stp_cm3_unit": "(cm^3_STP/cm^3)/(molec/uc)",
        "partial_pressure": 250000.0,
        "partial_pressure_unit": "Pa",
        "partial_fugacity": 121261.37081517381,
        "partial_fugacity_unit": "Pa",
        "adsorbate_density_average": 0.43238,
        "adsorbate_density_unit": "kg/m^3",
        "adsorbate_density_dev": 0.23112,
        "enthalpy_of_adsorption_average": 740.33952,
        "enthalpy_of_adsorption_unit": "K",
        "enthalpy_of_adsorption_dev": 398.536954,
        "loading_absolute_average": 0.07,
        "loading_absolute_dev": 0.0374165739,
        "loading_absolute_unit": "molecules/unit cell",
        "loading_excess_average": 0.07,
        "loading_excess_dev": 0.0374165739,
        "loading_excess_unit": "molecules/unit cell",
        "widom_rosenbluth_factor_unit": "-",
        "widom_rosenbluth_factor_dev": None,
        "widom_rosenbluth_factor_average": None,
        "chemical_potential_unit": "K",
        "chemical_potential_dev": None,
        "chemical_potential_average": None,
        "henry_coefficient_unit": "mol/kg/Pa",
        "henry_coefficient_dev": None,
        "henry_coefficient_average": None,
        "adsorption_energy_widom_unit": "kJ/mol",
        "adsorption_energy_widom_dev": None,
        "adsorption_energy_widom_average": None,
    }

    propane = {
        "mol_fraction": 0.5,
        "mol_fraction_unit": "-",
        "conversion_factor_molec_uc_to_mol_kg": None,
        "conversion_factor_molec_uc_to_mol_kg_unit": "(mol/kg)/(molec/uc)",
        "conversion_factor_molec_uc_to_mg_g": None,
        "conversion_factor_molec_uc_to_mg_g_unit": "(mg/g)/(molec/uc)",
        "conversion_factor_molec_uc_to_cm3stp_gr": None,
        "conversion_factor_molec_uc_to_cm3stp_gr_unit": "(cm^3_STP/gr)/(molec/uc)",
        "conversion_factor_molec_uc_to_cm3stp_cm3": 2.3820335839,
        "conversion_factor_molec_uc_to_cm3stp_cm3_unit": "(cm^3_STP/cm^3)/(molec/uc)",
        "partial_pressure": 250000.0,
        "partial_pressure_unit": "Pa",
        "partial_fugacity": 230618.81828829306,
        "partial_fugacity_unit": "Pa",
        "adsorbate_density_average": 3.97159,
        "adsorbate_density_unit": "kg/m^3",
        "adsorbate_density_dev": 0.7045,
        "enthalpy_of_adsorption_average": 103.567,
        "enthalpy_of_adsorption_unit": "K",
        "enthalpy_of_adsorption_dev": 98.162438,
        "loading_absolute_average": 0.8475,
        "loading_absolute_dev": 0.1503329638,
        "loading_absolute_unit": "molecules/unit cell",
        "loading_excess_average": 0.8475,
        "loading_excess_dev": 0.1503329638,
        "loading_excess_unit": "molecules/unit cell",
        "widom_rosenbluth_factor_unit": "-",
        "widom_rosenbluth_factor_dev": None,
        "widom_rosenbluth_factor_average": None,
        "chemical_potential_unit": "K",
        "chemical_potential_dev": None,
        "chemical_potential_average": None,
        "henry_coefficient_unit": "mol/kg/Pa",
        "henry_coefficient_dev": None,
        "henry_coefficient_average": None,
        "adsorption_energy_widom_unit": "kJ/mol",
        "adsorption_energy_widom_dev": None,
        "adsorption_energy_widom_average": None,
    }

    for key, value in general.items():
        assert value == parsed_parameters["general"][key]

    for key, value in butane.items():
        assert value == parsed_parameters["components"]["butane"][key]

    for key, value in propane.items():
        assert value == parsed_parameters["components"]["propane"][key]


def test_parse_output_widom():
    """Testing output parser on an output file of a widom calculation of H2 in an empty COF"""

    with Path(CWD, "outputs/widom_insertion.out").open("r", encoding="utf-8") as handle:
        parsed_parameters = parse_base_output(handle.read(), system_name="system1", ncomponents=1)[0]

    general = {
        "exceeded_walltime": False,
        "framework_density": "935.523150616336",
        "framework_density_unit": "kg/m^3",
        "energy_unit": "kJ/mol",
        "energy_host/ads_tot_initial": 0.0,
        "energy_host/ads_vdw_initial": 0.0,
        "energy_host/ads_coulomb_initial": 0.0,
        "energy_host/ads_tot_final": 0.0,
        "energy_host/ads_vdw_final": 0.0,
        "energy_host/ads_coulomb_final": 0.0,
        "cell_volume_average": 56953.44,
        "cell_volume_unit": "A^3",
        "cell_volume_dev": 0.0,
        "box_ax_average": 46.009,
        "box_ax_unit": "A^3",
        "box_ax_dev": 0.0,
        "box_by_average": 46.01567,
        "box_by_unit": "A^3",
        "box_by_dev": 0.0,
        "box_cz_average": 26.90119,
        "box_cz_unit": "A^3",
        "box_cz_dev": 0.0,
        "box_alpha_average": 121.11,
        "box_alpha_unit": "degrees",
        "box_alpha_dev": 0.0,
        "box_beta_average": 89.6387,
        "box_beta_unit": "degrees",
        "box_beta_dev": 0.0,
        "box_gamma_average": 90.2479,
        "box_gamma_unit": "degrees",
        "box_gamma_dev": 0.0,
        "adsorbate_density_average": 0.0,
        "adsorbate_density_unit": "kg/m^3",
        "adsorbate_density_dev": 0.0,
        "enthalpy_of_adsorption_average": None,
        "enthalpy_of_adsorption_unit": "K",
        "enthalpy_of_adsorption_dev": None,
        "energy_ads/ads_tot_average": 0.0,
        "energy_ads/ads_vdw_average": 0.0,
        "energy_ads/ads_coulomb_average": 0.0,
        "energy_ads/ads_tot_dev": 0.0,
        "energy_ads/ads_vdw_dev": 0.0,
        "energy_ads/ads_coulomb_dev": 0.0,
        "energy_host/ads_tot_average": 0.0,
        "energy_host/ads_vdw_average": 0.0,
        "energy_host/ads_coulomb_average": 0.0,
        "energy_host/ads_tot_dev": 0.0,
        "energy_host/ads_vdw_dev": 0.0,
        "energy_host/ads_coulomb_dev": 0.0,
        "tail_correction_energy_average": -37375.15068,
        "tail_correction_energy_unit": "K",
        "tail_correction_energy_dev": 0.0,
    }

    hydrogen = {
        "mol_fraction": 1.0,
        "mol_fraction_unit": "-",
        "conversion_factor_molec_uc_to_mol_kg": 0.1246621177,
        "conversion_factor_molec_uc_to_mol_kg_unit": "(mol/kg)/(molec/uc)",
        "conversion_factor_molec_uc_to_mg_g": 0.2493242354,
        "conversion_factor_molec_uc_to_mg_g_unit": "(mg/g)/(molec/uc)",
        "conversion_factor_molec_uc_to_cm3stp_gr": 2.7941736824,
        "conversion_factor_molec_uc_to_cm3stp_gr_unit": "(cm^3_STP/gr)/(molec/uc)",
        "conversion_factor_molec_uc_to_cm3stp_cm3": 2.6140141668,
        "conversion_factor_molec_uc_to_cm3stp_cm3_unit": "(cm^3_STP/cm^3)/(molec/uc)",
        "partial_pressure": -16605402.0,
        "partial_pressure_unit": "Pa",
        "partial_fugacity": 16605402.0,
        "partial_fugacity_unit": "Pa",
        "adsorbate_density_average": 0.0,
        "adsorbate_density_unit": "kg/m^3",
        "adsorbate_density_dev": 0.0,
        "loading_absolute_average": 0.0,
        "loading_absolute_dev": 0.0,
        "loading_absolute_unit": "molecules/unit cell",
        "loading_excess_average": 0.0,
        "loading_excess_dev": 0.0,
        "loading_excess_unit": "molecules/unit cell",
        "widom_rosenbluth_factor_unit": "-",
        "widom_rosenbluth_factor_dev": 78741.06669,
        "widom_rosenbluth_factor_average": 6973260.0,
        "chemical_potential_unit": "K",
        "chemical_potential_dev": 0.86752,
        "chemical_potential_average": -2056.48,
        "henry_coefficient_unit": "mol/kg/Pa",
        "henry_coefficient_dev": 0.131469,
        "henry_coefficient_average": 11.6428,
        "adsorption_energy_widom_unit": "kJ/mol",
        "adsorption_energy_widom_dev": 0.0111060559,
        "adsorption_energy_widom_average": -13.5545306415,
    }

    for key, value in general.items():
        assert value == parsed_parameters["general"][key]

    for key, value in hydrogen.items():
        assert value == parsed_parameters["components"]["H2"][key]



"""Run RASPA calculation with components mixture."""

import sys

import click
from aiida.common import NotExistent
from aiida.engine import run, run_get_pk
from aiida.orm import Code, Dict


def example_binary_misture(raspa_code, submit=True):
    """Prepare and submit RASPA calculation with components mixture."""

    # parameters
    parameters = Dict(
        dict={
            "GeneralSettings": {
                "SimulationType": "MonteCarlo",
                "NumberOfCycles": 50,
                "NumberOfInitializationCycles": 50,
                "PrintEvery": 10,
                "Forcefield": "GenericMOFs",
                "EwaldPrecision": 1e-6,
                "CutOff": 12.0,
            },
            "System": {
                "box_25_angstrom": {
                    "type": "Box",
                    "BoxLengths": "25 25 25",
                    "ExternalTemperature": 300.0,
                    "ExternalPressure": 5e5,
                },
            },
            "Component": {
                "propane": {
                    "MoleculeDefinition": "TraPPE",
                    "TranslationProbability": 1.0,
                    "RotationProbability": 1.0,
                    "ReinsertionProbability": 1.0,
                    "SwapProbability": 1.0,
                    "CreateNumberOfMolecules": 30,
                },
                "butane": {
                    "MoleculeDefinition": "TraPPE",
                    "TranslationProbability": 1.0,
                    "RotationProbability": 1.0,
                    "ReinsertionProbability": 1.0,
                    "SwapProbability": 1.0,
                    "CreateNumberOfMolecules": 30,
                },
            },
        }
    )

    # Contructing builder
    builder = raspa_code.get_builder()
    builder.parameters = parameters
    builder.metadata.options = {
        "resources": {
            "num_machines": 1,
            "num_mpiprocs_per_machine": 1,
        },
        "max_wallclock_seconds": 1 * 30 * 60,  # 30 min
        "withmpi": False,
    }
    builder.metadata.dry_run = False
    builder.metadata.store_provenance = True

    if submit:
        print("Testing RASPA with binary mixture (propane/butane) ...")
        res, pk = run_get_pk(builder)
        print("calculation pk: ", pk)
        print(
            "Average number of propane molecules/uc:",
            res["output_parameters"].dict.box_25_angstrom["components"]["propane"]["loading_absolute_average"],
        )
        print("OK, calculation has completed successfully")
    else:
        print("Generating test input ...")
        builder.metadata.dry_run = True
        builder.metadata.store_provenance = False
        run(builder)
        print("Submission test successful")
        print("In order to actually submit, add '--submit'")
    print("-----")


@click.command("cli")
@click.argument("codelabel")
@click.option("--submit", is_flag=True, help="Actually submit calculation")
def cli(codelabel, submit):
    """Click interface"""
    try:
        code = Code.get_from_string(codelabel)
    except NotExistent:
        print(f"The code '{codelabel}' does not exist")
        sys.exit(1)
    example_binary_misture(code, submit)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter

# EOF



"""Two-component GCMC through RaspaBaseWorkChain"""

import os
import sys

import click
from aiida.common import NotExistent
from aiida.engine import run_get_node
from aiida.orm import CifData, Code, Dict, SinglefileData

from aiida_raspa.workchains import RaspaBaseWorkChain


def example_base_workchain_gcmc(raspa_code):
    """Run base workchain for GCMC calculations with 2 components."""

    # pylint: disable=no-member

    print("Testing RASPA Xenon:Krypton GCMC through RaspaBaseWorkChain ...")

    parameters = Dict(
        dict={
            "GeneralSettings": {
                "SimulationType": "MonteCarlo",
                "NumberOfCycles": 50,
                "NumberOfInitializationCycles": 50,
                "PrintEvery": 10,
                "Forcefield": "GenericMOFs",
                "RemoveAtomNumberCodeFromLabel": True,
                "EwaldPrecision": 1e-6,
                "CutOff": 12.0,
            },
            "System": {
                "irmof_1": {
                    "type": "Framework",
                    "UnitCells": "1 1 1",
                    "HeliumVoidFraction": 0.149,
                    "ExternalTemperature": 300.0,
                    "ExternalPressure": 1e5,
                }
            },
            "Component": {
                "krypton": {
                    "MoleculeDefinition": "TraPPE",
                    "TranslationProbability": 0.5,
                    "ReinsertionProbability": 0.5,
                    "SwapProbability": 1.0,
                    "BlockPocketsFileName": {
                        "irmof_1": "irmof_1_krypton",
                    },
                },
                "xenon": {
                    "MoleculeDefinition": "TraPPE",
                    "TranslationProbability": 0.5,
                    "ReinsertionProbability": 0.5,
                    "SwapProbability": 1.0,
                    "BlockPocketsFileName": {
                        "irmof_1": "irmof_1_xenon",
                    },
                },
            },
        }
    )

    # framework
    pwd = os.path.dirname(os.path.realpath(__file__))
    structure = CifData(file=os.path.join(pwd, "..", "files", "IRMOF-1.cif"))
    structure_label = "irmof_1"

    block_pocket_node1 = SinglefileData(file=os.path.join(pwd, "..", "files", "IRMOF-1_test.block")).store()
    block_pocket_node2 = SinglefileData(file=os.path.join(pwd, "..", "files", "IRMOF-1_test.block")).store()

    # Constructing builder
    builder = RaspaBaseWorkChain.get_builder()

    # Specifying the code
    builder.raspa.code = raspa_code

    # Specifying the framework
    builder.raspa.framework = {
        structure_label: structure,
    }

    # Specifying the input parameters
    builder.raspa.parameters = parameters

    # Specifying the block pockets
    builder.raspa.block_pocket = {
        "irmof_1_krypton": block_pocket_node1,
        "irmof_1_xenon": block_pocket_node2,
    }

    builder.handler_overrides = Dict(
        dict={"check_gcmc_convergence": True}
    )  # Enable gcmc convergence handler disabled by default.

    # Specifying the scheduler options
    builder.raspa.metadata.options = {
        "resources": {
            "num_machines": 1,
            "num_mpiprocs_per_machine": 1,
        },
        "max_wallclock_seconds": 1 * 30 * 60,  # 30 min
        "withmpi": False,
    }

    _, node = run_get_node(builder)
    assert node.exit_status == 0


@click.command("cli")
@click.argument("codelabel")
def cli(codelabel):
    """Click interface"""
    try:
        code = Code.get_from_string(codelabel)
    except NotExistent:
        print(f"The code '{codelabel}' does not exist")
        sys.exit(1)
    example_base_workchain_gcmc(code)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter

# EOF



"""RASPA inspection tools"""
from aiida.engine import calcfunction
from aiida.orm import Dict


@calcfunction
def add_write_binary_restart(input_dict, write_every):
    final_dict = input_dict.get_dict()
    final_dict["GeneralSettings"]["WriteBinaryRestartFileEvery"] = write_every
    return input_dict if input_dict.get_dict() == final_dict else Dict(final_dict)


def modify_number_of_cycles(input_dict, additional_init_cycle, additional_prod_cycle):
    """Modify number of cycles to improve the convergence."""
    final_dict = input_dict.get_dict()
    try:
        final_dict["GeneralSettings"]["NumberOfInitializationCycles"] += additional_init_cycle.value
    except KeyError:
        final_dict["GeneralSettings"]["NumberOfInitializationCycles"] = additional_init_cycle.value

    final_dict["GeneralSettings"]["NumberOfCycles"] += additional_prod_cycle.value

    # It is a restart job the number of molecules need to be set to zero
    for component in final_dict["Component"].values():
        if "CreateNumberOfMolecules" in component:
            if isinstance(component["CreateNumberOfMolecules"], dict):
                for number in component["CreateNumberOfMolecules"]:
                    component["CreateNumberOfMolecules"][number] = 0
            else:
                component["CreateNumberOfMolecules"] = 0

    # Return final_dict dict only if it was modified
    return input_dict if input_dict.get_dict() == final_dict else Dict(final_dict)


@calcfunction
def increase_box_lenght(input_dict, box_name, box_length_current):
    """Increase the box lenght to improve the convegence."""
    import math

    final_dict = input_dict.get_dict()
    bx_length_old = [float(element) for element in final_dict["System"][box_name.value]["BoxLengths"].split()]

    # We do the simulation in cubic box.
    addition_length = abs(math.ceil(bx_length_old[0] - box_length_current.value))
    box_one_length_new = [bx_length_old[i] + addition_length for i in range(3)]
    final_dict["System"][box_name.value]["BoxLengths"] = " ".join(str(l) for l in box_one_length_new)

    return Dict(final_dict)



"""
For pytest initialise a test database and profile
"""
import pytest

pytest_plugins = ["aiida.manage.tests.pytest_fixtures"]  # pylint: disable=invalid-name


@pytest.fixture(scope="function")
def raspa_code(aiida_local_code_factory):  # pylint: disable=unused-argument
    return aiida_local_code_factory("raspa", "simulate")



"""Base work chain to run a RASPA calculation"""

from aiida.common import AttributeDict
from aiida.engine import (
    BaseRestartWorkChain,
    ProcessHandlerReport,
    process_handler,
    while_,
)
from aiida.orm import Float, Int, Str
from aiida.plugins import CalculationFactory

from aiida_raspa.utils import (
    add_write_binary_restart,
    increase_box_lenght,
    modify_number_of_cycles,
)

RaspaCalculation = CalculationFactory("raspa")  # pylint: disable=invalid-name


class RaspaBaseWorkChain(BaseRestartWorkChain):
    """Workchain to run a RASPA calculation with automated error handling and restarts."""

    _process_class = RaspaCalculation

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(RaspaCalculation, namespace="raspa")
        spec.outline(
            cls.setup,
            while_(cls.should_run_process)(
                cls.run_process,
                cls.inspect_process,
            ),
            cls.results,
        )
        spec.expose_outputs(RaspaCalculation)

    def setup(self):
        """Call the `setup` of the `BaseRestartWorkChain` and then create the inputs dictionary in `self.ctx.inputs`.
        This `self.ctx.inputs` dictionary will be used by the `BaseRestartWorkChain` to submit the calculations in the
        internal loop."""

        super().setup()

        self.ctx.inputs = AttributeDict(self.exposed_inputs(RaspaCalculation, "raspa"))
        if "WriteBinaryRestartFileEvery" not in self.ctx.inputs.parameters["GeneralSettings"]:
            self.ctx.inputs.parameters = add_write_binary_restart(self.ctx.inputs.parameters, Int(1000))

    def report_error_handled(self, calculation, action):
        """Report an action taken for a calculation that has failed.
        This should be called in a registered error handler if its condition is met and an action was taken.
        :param calculation: the failed calculation node
        :param action: a string message with the action taken"""

        arguments = [calculation.process_label, calculation.pk, calculation.exit_status, calculation.exit_message]
        self.report("{}<{}> failed with exit status {}: {}".format(*arguments))
        self.report(f"Action taken: {action}")

    @process_handler(priority=570, exit_codes=RaspaCalculation.exit_codes.TIMEOUT, enabled=True)
    def handle_timeout(self, calculation):
        """Error handler that restarts calculation finished with TIMEOUT ExitCode."""
        self.report_error_handled(calculation, "Timeout handler. Adding remote folder as input to use binary restart.")
        self.ctx.inputs.parent_folder = calculation.outputs.remote_folder
        return ProcessHandlerReport(False)

    @process_handler(priority=400, enabled=False)
    def check_widom_convergence(self, calculation):
        """Checks whether a Widom particle insertion is converged. The check is based on the
        error bar of the Henry coefficient."""

        conv_threshold = 0.1
        additional_cycle = 2000

        output_widom = calculation.outputs.output_parameters.get_dict()
        structure_label = list(calculation.get_incoming().nested()["framework"].keys())[0]
        conv_stat = []

        for comp in calculation.inputs.parameters["Component"]:
            kh_average_comp = output_widom[structure_label]["components"][comp]["henry_coefficient_average"]
            kh_dev_comp = output_widom[structure_label]["components"][comp]["henry_coefficient_dev"]

            error = round((kh_dev_comp / kh_average_comp), 2)
            if error <= conv_threshold:
                conv_stat.append(True)
            else:
                conv_stat.append(False)

        if not all(conv_stat):
            self.report("Widom particle insertion calculationulation is NOT converged: repeating with more trials...")
            self.ctx.inputs.retrieved_parent_folder = calculation.outputs["retrieved"]
            self.ctx.inputs.parameters = modify_number_of_cycles(
                self.ctx.inputs.parameters, additional_init_cycle=Int(0), additional_prod_cycle=Int(additional_cycle)
            )
            return ProcessHandlerReport(False)

        return None

    @process_handler(priority=410, enabled=False)
    def check_gcmc_convergence(self, calc):
        """Checks whether a GCMC calc is converged. Checking is based on the error bar on average loading."""
        conv_threshold = 0.1
        additional_init_cycle = 2000
        additional_prod_cycle = 2000

        output_gcmc = calc.outputs.output_parameters.get_dict()
        structure_label = list(calc.get_incoming().nested()["framework"].keys())[0]
        conv_stat = []

        for comp in calc.inputs.parameters["Component"]:
            loading_average_comp = output_gcmc[structure_label]["components"][comp]["loading_absolute_average"]
            loading_dev_comp = output_gcmc[structure_label]["components"][comp]["loading_absolute_dev"]

            # It can happen for weekly adsorbed species.
            # we need to think about a better way to handle it.
            # Currently, if it happens for five iterations, self will not continue.
            if loading_average_comp == 0:
                conv_stat.append(False)
            else:
                error = round((loading_dev_comp / loading_average_comp), 2)
                if error <= conv_threshold:
                    conv_stat.append(True)
                else:
                    conv_stat.append(False)

        if not all(conv_stat):
            self.report("GCMC calculation is NOT converged: continuing from restart...")
            self.ctx.inputs.retrieved_parent_folder = calc.outputs["retrieved"]
            self.ctx.inputs.parameters = modify_number_of_cycles(
                self.ctx.inputs.parameters,
                additional_init_cycle=Int(additional_init_cycle),
                additional_prod_cycle=Int(additional_prod_cycle),
            )
            return ProcessHandlerReport(False)

        return None

    @process_handler(priority=410, enabled=False)
    def check_gemc_convergence(self, calc):
        """Checks whether a GEMC calc is converged. Checking is based on the error bar on average loading which is
        average number of molecules in each simulation box."""

        conv_threshold = 0.1
        additional_init_cycle = 2000
        additional_prod_cycle = 2000

        output_gemc = calc.outputs.output_parameters.get_dict()
        conv_stat = []

        for comp in calc.inputs.parameters["Component"]:
            molec_per_box1_comp_average = output_gemc["box_one"]["components"][comp]["loading_absolute_average"]
            molec_per_box2_comp_average = output_gemc["box_two"]["components"][comp]["loading_absolute_average"]
            molec_per_box1_comp_dev = output_gemc["box_one"]["components"][comp]["loading_absolute_dev"]
            molec_per_box2_comp_dev = output_gemc["box_two"]["components"][comp]["loading_absolute_dev"]

            error_box1 = round((molec_per_box1_comp_dev / molec_per_box1_comp_average), 2)
            error_box2 = round((molec_per_box2_comp_dev / molec_per_box2_comp_average), 2)

            if (error_box1 <= conv_threshold) and (error_box2 <= conv_threshold):
                conv_stat.append(True)
            else:
                conv_stat.append(False)

        if not all(conv_stat):
            self.report("GEMC calculation is NOT converged: continuing from restart...")
            self.ctx.inputs.retrieved_parent_folder = calc.outputs["retrieved"]
            self.ctx.inputs.parameters = modify_number_of_cycles(
                self.ctx.inputs.parameters,
                additional_init_cycle=Int(additional_init_cycle),
                additional_prod_cycle=Int(additional_prod_cycle),
            )
            return ProcessHandlerReport(False)

        return None

    @process_handler(priority=400, enabled=False)
    def check_gemc_box(self, calc):
        """Checks whether each simulation box still satisfies minimum image convention."""

        output_gemc = calc.outputs.output_parameters.get_dict()
        cutoff = calc.inputs.parameters["GeneralSettings"]["CutOff"]
        box_one_stat = []
        box_two_stat = []

        box_one_length_current = []
        box_two_length_current = []

        for box_len_ave in ["box_ax_average", "box_by_average", "box_cz_average"]:
            if output_gemc["box_one"]["general"][box_len_ave] > 2 * cutoff:
                box_one_stat.append(True)
            else:
                box_one_stat.append(False)
                box_one_length_current.append(output_gemc["box_one"]["general"][box_len_ave])

            if output_gemc["box_two"]["general"][box_len_ave] > 2 * cutoff:
                box_two_stat.append(True)
            else:
                box_two_stat.append(False)
                box_two_length_current.append(output_gemc["box_two"]["general"][box_len_ave])

        if not all(box_one_stat and box_two_stat):
            self.report("GEMC box is NOT converged: repeating with increase box...")
            # Fixing the issue.
            if not all(box_one_stat):
                self.ctx.inputs.parameters = increase_box_lenght(
                    self.ctx.inputs.parameters, Str("box_one"), Float(box_one_length_current[0])
                )

            if not all(box_two_stat):
                self.ctx.inputs.parameters = increase_box_lenght(
                    self.ctx.inputs.parameters, Str("box_two"), Float(box_two_length_current[0])
                )

            return ProcessHandlerReport(True)

        return None



""" Sphinx configuration for aiida-raspa

This file is execfile()d with the current directory set to its
containing dir.

Note that not all possible configuration values are present in this
autogenerated file.

All configuration values have a default; values that are commented out
serve to show the default.
"""

import os
import sys
import time

from aiida.manage.configuration import load_documentation_profile

# pylint: disable=invalid-name,ungrouped-imports

# -- AiiDA-related setup --------------------------------------------------

# Load the dummy profile even if we are running locally, this way the documentation will succeed even if the current
# default profile of the AiiDA installation does not use a Django backend.
load_documentation_profile()

# If we are not on READTHEDOCS load the Sphinx theme manually
if not os.environ.get("READTHEDOCS", None):
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]


import aiida_raspa  # pylint: disable=wrong-import-position

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "1.5"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinxcontrib.contentui",
    "aiida.sphinxext",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/2.7", None),
    "aiida": ("https://aiida.readthedocs.io/projects/aiida-core/en/latest/", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
# ~ master_doc = 'index'
master_doc = "index"

# General information about the project.
project = "aiida-raspa"
copyright_first_year = "2019"
copyright_owners = "The AiiDA Team"

current_year = str(time.localtime().tm_year)
copyright_year_string = (
    current_year if current_year == copyright_first_year else f"{copyright_first_year}-{current_year}"
)
# pylint: disable=redefined-builtin
copyright = f"{copyright_year_string}, {copyright_owners}. All rights reserved"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = aiida_raspa.__version__
# The short X.Y version.
version = ".".join(release.split(".")[:2])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# exclude_patterns = ['doc.rst']
# ~ exclude_patterns = ['index.rst']

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# ~ html_theme = 'basicstrap'
## SET BELOW

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# ~ html_theme_options = {
# ~ 'inner_theme': True,
# ~ 'inner_theme_name': 'bootswatch-darkly',
# ~ 'nav_fixed_top': False
# ~ }

# Add any paths that contain custom themes here, relative to this directory.
# ~ html_theme_path = ["."]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = "images/.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = "images/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# ~ html_show_copyright = False

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
html_use_opensearch = "http://aiida-raspa.readthedocs.io"

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
html_search_language = "en"

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
# html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = "aiida-raspa-doc"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #'preamble': '',
    # Latex figure (float) alignment
    #'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
# latex_documents = [
# ]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


def run_apidoc(_):
    """Runs sphinx-apidoc when building the documentation.

    Needs to be done in conf.py in order to include the APIdoc in the
    build on readthedocs.

    See also https://github.com/rtfd/readthedocs.org/issues/1139
    """
    source_dir = os.path.abspath(os.path.dirname(__file__))
    apidoc_dir = os.path.join(source_dir, "apidoc")
    package_dir = os.path.join(source_dir, os.pardir, os.pardir, "aiida_raspa")

    # In #1139, they suggest the route below, but this ended up
    # calling sphinx-build, not sphinx-apidoc
    # from sphinx.apidoc import main
    # main([None, '-e', '-o', apidoc_dir, package_dir, '--force'])

    import subprocess

    cmd_path = "sphinx-apidoc"
    if hasattr(sys, "real_prefix"):  # Check to see if we are in a virtualenv
        # If we are, assemble the path manually
        cmd_path = os.path.abspath(os.path.join(sys.prefix, "bin", "sphinx-apidoc"))

    options = [
        "-o",
        apidoc_dir,
        package_dir,
        "--private",
        "--force",
        "--no-toc",
    ]

    # See https://stackoverflow.com/a/30144019
    env = os.environ.copy()
    env["SPHINX_APIDOC_OPTIONS"] = "members,special-members,private-members,undoc-members,show-inheritance"
    subprocess.check_call([cmd_path] + options, env=env)


def setup(app):
    """Set up documentation."""
    app.connect("builder-inited", run_apidoc)


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# man_pages = [
# ]

# If true, show URL addresses after external links.
# man_show_urls = False

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
# texinfo_documents = [
# ]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False

# Warnings to ignore when using the -n (nitpicky) option
# We should ignore any python built-in exception, for instance
nitpick_ignore = [
    ("py:exc", "ArithmeticError"),
    ("py:exc", "AssertionError"),
    ("py:exc", "AttributeError"),
    ("py:exc", "BaseException"),
    ("py:exc", "BufferError"),
    ("py:exc", "DeprecationWarning"),
    ("py:exc", "EOFError"),
    ("py:exc", "EnvironmentError"),
    ("py:exc", "Exception"),
    ("py:exc", "FloatingPointError"),
    ("py:exc", "FutureWarning"),
    ("py:exc", "GeneratorExit"),
    ("py:exc", "IOError"),
    ("py:exc", "ImportError"),
    ("py:exc", "ImportWarning"),
    ("py:exc", "IndentationError"),
    ("py:exc", "IndexError"),
    ("py:exc", "KeyError"),
    ("py:exc", "KeyboardInterrupt"),
    ("py:exc", "LookupError"),
    ("py:exc", "MemoryError"),
    ("py:exc", "NameError"),
    ("py:exc", "NotImplementedError"),
    ("py:exc", "OSError"),
    ("py:exc", "OverflowError"),
    ("py:exc", "PendingDeprecationWarning"),
    ("py:exc", "ReferenceError"),
    ("py:exc", "RuntimeError"),
    ("py:exc", "RuntimeWarning"),
    ("py:exc", "StandardError"),
    ("py:exc", "StopIteration"),
    ("py:exc", "SyntaxError"),
    ("py:exc", "SyntaxWarning"),
    ("py:exc", "SystemError"),
    ("py:exc", "SystemExit"),
    ("py:exc", "TabError"),
    ("py:exc", "TypeError"),
    ("py:exc", "UnboundLocalError"),
    ("py:exc", "UnicodeDecodeError"),
    ("py:exc", "UnicodeEncodeError"),
    ("py:exc", "UnicodeError"),
    ("py:exc", "UnicodeTranslateError"),
    ("py:exc", "UnicodeWarning"),
    ("py:exc", "UserWarning"),
    ("py:exc", "VMSError"),
    ("py:exc", "ValueError"),
    ("py:exc", "Warning"),
    ("py:exc", "WindowsError"),
    ("py:exc", "ZeroDivisionError"),
    ("py:obj", "str"),
    ("py:obj", "list"),
    ("py:obj", "tuple"),
    ("py:obj", "int"),
    ("py:obj", "float"),
    ("py:obj", "bool"),
    ("py:obj", "Mapping"),
    ("py:class", "list"),
    ("py:class", "tuple"),
    ("py:class", "dict"),
]



"""Raspa input plugin."""
import os
from pathlib import Path
from shutil import copyfile, copytree

from aiida.common import CalcInfo, CodeInfo, InputValidationError

# from aiida.cmdline.utils import echo
from aiida.engine import CalcJob
from aiida.orm import Dict, FolderData, List, RemoteData, SinglefileData
from aiida.plugins import DataFactory

from aiida_raspa.utils import RaspaInput

# data objects
CifData = DataFactory("core.cif")  # pylint: disable=invalid-name


class RaspaCalculation(CalcJob):
    """This is a RaspaCalculation, subclass of CalcJob, to prepare input for RASPA code.
    For information on RASPA, refer to: https://github.com/iraspa/raspa2.
    """

    # Defaults
    INPUT_FILE = "simulation.input"
    OUTPUT_FOLDER = "Output"
    RESTART_FOLDER = "Restart"
    PROJECT_NAME = "aiida"
    DEFAULT_PARSER = "raspa"

    @classmethod
    def define(cls, spec):
        super().define(spec)

        # Input parameters
        spec.input("parameters", valid_type=Dict, required=True, help="Input parameters")
        spec.input_namespace("framework", valid_type=CifData, required=False, dynamic=True, help="Input framework(s)")
        spec.input_namespace(
            "block_pocket", valid_type=SinglefileData, required=False, dynamic=True, help="Zeo++ block pocket file"
        )
        spec.input_namespace(
            "file", valid_type=SinglefileData, required=False, dynamic=True, help="Additional input file(s)"
        )
        spec.input("settings", valid_type=Dict, required=False, help="Additional input parameters")
        spec.input(
            "parent_folder",
            valid_type=RemoteData,
            required=False,
            help="Remote folder used to continue the same simulation stating from the binary restarts.",
        )
        spec.input(
            "retrieved_parent_folder",
            valid_type=FolderData,
            required=False,
            help="To use an old calculation as a starting poing for a new one.",
            validator=cls.validate_retrieved_parent_folder,
        )
        spec.inputs["metadata"]["options"]["parser_name"].default = cls.DEFAULT_PARSER
        spec.inputs["metadata"]["options"]["resources"].default = {
            "num_machines": 1,
            "num_mpiprocs_per_machine": 1,
            "num_cores_per_mpiproc": 1,
        }
        spec.inputs["metadata"]["options"]["withmpi"].default = False

        # Output parameters
        spec.output("output_parameters", valid_type=Dict, required=True, help="The results of a calculation")
        spec.output("warnings", valid_type=List, required=False, help="Warnings that appeared during the calculation")

        # Exit codes
        spec.exit_code(
            100, "ERROR_NO_RETRIEVED_FOLDER", message="The retrieved folder data node could not be accessed."
        )
        spec.exit_code(101, "ERROR_NO_OUTPUT_FILE", message="The retrieved folder does not contain an output file.")
        spec.exit_code(
            102, "ERROR_SIMULATION_DID_NOT_START", message='The output does not contain "Starting simulation".'
        )
        spec.exit_code(500, "TIMEOUT", message="The calculation could not be completed due to the lack of time.")

        # Default output node
        spec.default_output_node = "output_parameters"

    @staticmethod
    def validate_retrieved_parent_folder(value, _):
        """Validate the `retrieved_parent_folder` input."""

        repository = value.base.repository

        if "Restart" not in repository.list_object_names():
            return "Restart was requested but the restart folder was not found in the previous calculation."
        for system_path in repository.list_object_names("Restart"):
            if len(repository.list_object_names(Path("Restart") / system_path)) != 1:
                return "There was more than one file in a system directory of the `Restart` directory."

    # --------------------------------------------------------------------------
    # pylint: disable = too-many-locals
    def prepare_for_submission(self, folder):
        """
        This is the routine to be called when you want to create
        the input files and related stuff with a plugin.

        :param folder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        """
        # create calc info
        calcinfo = CalcInfo()
        calcinfo.remote_copy_list = []
        calcinfo.local_copy_list = []

        # initialize input parameters
        inp = RaspaInput(self.inputs.parameters.get_dict())

        # keep order of systems in the extras
        self.node.set_extra("system_order", inp.system_order)

        # handle framework(s) and/or box(es)
        if "System" in inp.params:
            self._handle_system_section(inp.params["System"], folder)

        # handle restart
        if "retrieved_parent_folder" in self.inputs:
            calcinfo.local_copy_list.extend(self._handle_retrieved_parent_folder(inp))
            inp.params["GeneralSettings"]["RestartFile"] = True

        # handle binary restart
        if "parent_folder" in self.inputs:
            inp.params["GeneralSettings"]["ContinueAfterCrash"] = True
            calcinfo.remote_copy_list.append(
                (
                    self.inputs.parent_folder.computer.uuid,
                    os.path.join(self.inputs.parent_folder.get_remote_path(), "CrashRestart"),
                    "CrashRestart",
                )
            )

        # get settings
        if "settings" in self.inputs:
            settings = self.inputs.settings.get_dict()
        else:
            settings = {}

        # write raspa input file
        with open(folder.get_abs_path(self.INPUT_FILE), "w", encoding="utf-8") as fobj:
            fobj.write(inp.render())

        # create code info
        codeinfo = CodeInfo()
        codeinfo.cmdline_params = settings.pop("cmdline", []) + [self.INPUT_FILE]
        codeinfo.code_uuid = self.inputs.code.uuid

        calcinfo.stdin_name = self.INPUT_FILE
        calcinfo.uuid = self.uuid
        calcinfo.cmdline_params = codeinfo.cmdline_params
        calcinfo.stdin_name = self.INPUT_FILE
        # calcinfo.stdout_name = self.OUTPUT_FILE
        calcinfo.codes_info = [codeinfo]

        # file lists
        if "file" in self.inputs:
            for fobj in self.inputs.file.values():
                calcinfo.local_copy_list.append((fobj.uuid, fobj.filename, fobj.filename))

        # block pockets
        if "block_pocket" in self.inputs:
            for name, fobj in self.inputs.block_pocket.items():
                calcinfo.local_copy_list.append((fobj.uuid, fobj.filename, name + ".block"))

        calcinfo.retrieve_list = [self.OUTPUT_FOLDER, self.RESTART_FOLDER]
        calcinfo.retrieve_list += settings.pop("additional_retrieve_list", [])

        # check for left over settings
        if settings:
            raise InputValidationError(
                "The following keys have been found "
                + f"in the settings input node {self.pk}, "
                + "but were not understood: "
                + ",".join(list(settings.keys()))
            )

        return calcinfo

    def _handle_system_section(self, system_dict, folder):
        """Handle framework(s) and/or box(es)."""
        for name, sparams in system_dict.items():
            if sparams["type"] == "Framework":
                try:
                    self.inputs.framework[name].export(folder.get_abs_path(name + ".cif"), fileformat="cif")
                except KeyError as err:
                    raise InputValidationError(
                        f"You specified '{name}' framework in the input dictionary, but did not provide the input "
                        "framework with the same name"
                    ) from err

    def _handle_retrieved_parent_folder(self, inp):
        """Construct the local copy list from a the `retrieved_parent_folder` input."""
        local_copy_list = []

        parent_folder = self.inputs.retrieved_parent_folder
        base_src_path = Path("Restart")
        base_dest_path = Path("RestartInitial")

        for i_system, system_name in enumerate(inp.system_order):
            system = inp.params["System"][system_name]
            system_dir = f"System_{i_system}"

            old_fname = parent_folder.base.repository.list_object_names(base_src_path / system_dir).pop()

            if system["type"] == "Box":
                system_or_box = "Box"
                (n_x, n_y, n_z) = (1, 1, 1)
                if "ExternalPressure" not in system:
                    system["ExternalPressure"] = 0
            elif system["type"] == "Framework":
                system_or_box = system_name
                try:
                    (n_x, n_y, n_z) = tuple(map(int, system["UnitCells"].split()))
                except KeyError:
                    (n_x, n_y, n_z) = 1, 1, 1

            external_pressure = system["ExternalPressure"] if "ExternalPressure" in system else 0

            new_fname = (
                f"restart_{system_or_box:s}_{n_x:}.{n_y:d}.{n_z:d}_{system['ExternalTemperature']:f}"
                f"_{external_pressure:g}"
            )

            local_copy_list.append(
                (
                    parent_folder.uuid,
                    Path(base_src_path, system_dir, old_fname).as_posix(),
                    Path(base_dest_path, system_dir, new_fname).as_posix(),
                )
            )

        return local_copy_list



"""Run RASPA single-component GEMC calculation -- Restart"""

import sys

import click
import pytest
from aiida.common import NotExistent
from aiida.engine import run, run_get_pk
from aiida.orm import Code, Dict, load_node


def example_gemc_single_comp(raspa_code, gemc_single_comp_calc_pk=None, submit=True):
    """Prepare and submit RASPA calculation with components mixture."""

    # This line is needed for tests only
    if gemc_single_comp_calc_pk is None:
        gemc_single_comp_calc_pk = pytest.gemc_single_comp_calc_pk  # pylint: disable=no-member

    # parameters
    parameters = Dict(
        dict={
            "GeneralSettings": {
                "SimulationType": "MonteCarlo",
                "NumberOfCycles": 50,
                "NumberOfInitializationCycles": 50,
                "PrintEvery": 10,
                "Forcefield": "GenericMOFs",
                "EwaldPrecision": 1e-6,
                "CutOff": 12.0,
                "GibbsVolumeChangeProbability": 0.1,
            },
            "System": {
                "box_one": {
                    "type": "Box",
                    "BoxLengths": "25 25 25",
                    "BoxAngles": "90 90 90",
                    "ExternalTemperature": 200.0,
                },
                "box_two": {
                    "type": "Box",
                    "BoxLengths": "25 25 25",
                    "BoxAngles": "90 90 90",
                    "ExternalTemperature": 200.0,
                },
            },
            "Component": {
                "methane": {
                    "MoleculeDefinition": "TraPPE",
                    "TranslationProbability": 1.0,
                    "ReinsertionProbability": 1.0,
                    "GibbsSwapProbability": 1.0,
                    "CreateNumberOfMolecules": {
                        "box_one": 50,
                        "box_two": 50,
                    },
                },
            },
        }
    )

    # restart file
    retrieved_parent_folder = load_node(gemc_single_comp_calc_pk).outputs.retrieved

    # Contructing builder
    builder = raspa_code.get_builder()
    builder.parameters = parameters
    builder.retrieved_parent_folder = retrieved_parent_folder
    builder.metadata.options = {
        "resources": {
            "num_machines": 1,
            "num_mpiprocs_per_machine": 1,
        },
        "max_wallclock_seconds": 1 * 30 * 60,  # 30 min
        "withmpi": False,
    }
    builder.metadata.dry_run = False
    builder.metadata.store_provenance = True

    if submit:
        print("Testing RASPA GEMC with methane (Restart)...")
        res, pk = run_get_pk(builder)
        print("calculation pk: ", pk)
        print(
            "Average number of methane molecules/uc (box_one):",
            res["output_parameters"].dict.box_one["components"]["methane"]["loading_absolute_average"],
        )
        print(
            "Average number of methane molecules/uc (box_two):",
            res["output_parameters"].dict.box_two["components"]["methane"]["loading_absolute_average"],
        )
        print("OK, calculation has completed successfully")
    else:
        print("Generating test input ...")
        builder.metadata.dry_run = True
        builder.metadata.store_provenance = False
        run(builder)
        print("Submission test successful")
        print("In order to actually submit, add '--submit'")
    print("-----")


@click.command("cli")
@click.argument("codelabel")
@click.option("--previous_calc", "-p", required=True, type=int, help="PK of example_framework_box.py calculation")
@click.option("--submit", is_flag=True, help="Actually submit calculation")
def cli(codelabel, previous_calc, submit):
    """Click interface"""
    try:
        code = Code.get_from_string(codelabel)
    except NotExistent:
        print(f"The code '{codelabel}' does not exist")
        sys.exit(1)
    example_gemc_single_comp(code, previous_calc, submit)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter

# EOF



"""Test Raspa input generator"""

import pytest

from aiida_raspa.utils import RaspaInput


def test_render_empty_dict():
    with pytest.raises(KeyError) as excinfo:
        RaspaInput({})
    assert "should contain the System subdictionary" in str(excinfo)


def test_render_wrong_system():
    """Test non-dictionary value for System"""
    inp_dict = {
        "System": 1,
    }
    with pytest.raises(AttributeError) as excinfo:
        RaspaInput(inp_dict)
    assert "of type dict" in str(excinfo)


def test_render_empty_system():
    """Test empy System dictionary"""
    inp_dict = {
        "System": {},
    }
    with pytest.raises(ValueError) as excinfo:
        RaspaInput(inp_dict)
    assert "should not be empty" in str(excinfo)


def test_render_base_input():
    """Test normal input dictionary"""
    inp_dict = {
        "GeneralSettings": {
            "SimulationType": "MonteCarlo",
            "NumberOfCycles": 400,
            "NumberOfInitializationCycles": 200,
            "PrintEvery": 200,
            "Forcefield": "GenericMOFs",
            "EwaldPrecision": 1e-6,
            "CutOff": 12.0,
            "WriteBinaryRestartFileEvery": 200,
        },
        "System": {
            "tcc1rs": {
                "type": "Framework",
                "UnitCells": "1 1 1",
                "HeliumVoidFraction": 0.149,
                "ExternalTemperature": 300.0,
                "ExternalPressure": 5e5,
            },
        },
        "Component": {
            "methane": {
                "MoleculeDefinition": "TraPPE",
                "TranslationProbability": 0.5,
                "ReinsertionProbability": 0.5,
                "SwapProbability": 1.0,
                "CreateNumberOfMolecules": 0,
            }
        },
    }
    inp = RaspaInput(inp_dict)
    assert (
        inp.render()
        == """!!! Generated by AiiDA !!!
CutOff 12.0
EwaldPrecision 1e-06
Forcefield GenericMOFs
NumberOfCycles 400
NumberOfInitializationCycles 200
PrintEvery 200
SimulationType MonteCarlo
WriteBinaryRestartFileEvery 200
Framework 0
   FrameworkName tcc1rs
   ExternalPressure 500000.0
   ExternalTemperature 300.0
   HeliumVoidFraction 0.149
   UnitCells 1 1 1
Component 0 MoleculeName methane
   CreateNumberOfMolecules 0
   MoleculeDefinition TraPPE
   ReinsertionProbability 0.5
   SwapProbability 1.0
   TranslationProbability 0.5
"""
    )



"""Raspa output parser."""
import os
from pathlib import Path

from aiida.common import NotExistent, OutputParsingError
from aiida.engine import ExitCode
from aiida.orm import Dict, List
from aiida.parsers.parser import Parser

from aiida_raspa.utils import parse_base_output

# parser
# --------------------------------------------------------------------------------------------


class RaspaParser(Parser):
    """Parse RASPA output"""

    # --------------------------------------------------------------------------
    def parse(self, **kwargs):  # pylint: disable=too-many-locals
        """Receives in input a dictionary of retrieved nodes. Does all the logic here."""
        try:
            out_folder = self.retrieved
        except NotExistent:
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER
        output_folder_name = self.node.process_class.OUTPUT_FOLDER

        if output_folder_name not in out_folder.base.repository.list_object_names():  # pylint: disable=protected-access
            return self.exit_codes.ERROR_NO_OUTPUT_FILE

        output_parameters = {}
        warnings = []
        ncomponents = len(self.node.inputs.parameters.get_dict()["Component"])
        for system_id, system_name in enumerate(self.node.get_extra("system_order")):
            # specify the name for the system
            output_dir = Path(output_folder_name) / f"System_{system_id}"
            output_filename = out_folder.base.repository.list_object_names(output_dir).pop()
            output_contents = out_folder.base.repository.get_object_content(output_dir / output_filename)

            # Check for possible errors
            if "Starting simulation" not in output_contents:
                return self.exit_codes.ERROR_SIMULATION_DID_NOT_START
            if "Simulation finished" not in output_contents:
                return self.exit_codes.TIMEOUT

            # parse output parameters and warnings
            parsed_parameters, parsed_warnings = parse_base_output(output_contents, system_name, ncomponents)
            output_parameters[system_name] = parsed_parameters
            warnings += parsed_warnings

        self.out("output_parameters", Dict(dict=output_parameters))
        self.out("warnings", List(list=warnings))

        return ExitCode(0)



"""Raspa utils."""
from .base_input_generator import RaspaInput
from .base_parser import parse_base_output
from .inspection_tools import (
    add_write_binary_restart,
    increase_box_lenght,
    modify_number_of_cycles,
)



"""Basic raspa input generator."""
from copy import deepcopy

ORDERED_ITEMS_COMPONENT_SECTION = [
    "NumberOfIdentityChanges",
    "IdentityChangesList",
]


class RaspaInput:  # pylint: disable=too-few-public-methods
    """Convert input dictionary into input file"""

    def __init__(self, params):
        """Construct a `RaspaInput` instance."""
        self.params = deepcopy(params)  # make sure the original object is not modified
        try:
            self.system_order = sorted(params["System"].keys())  # we sort the keys to keep the order in the input
        except KeyError as err:
            raise KeyError("The input dictionary should contain the System subdictionary.") from err
        except AttributeError as err:
            raise AttributeError("The System subdictionary should be of type dict.") from err
        # file persistent this is essential for the restarts.

        if not self.system_order:
            raise ValueError("The System subdictionary should not be empty.")

    # --------------------------------------------------------------------------
    def render(self):
        """Perform conversion"""
        output = ["!!! Generated by AiiDA !!!"]
        params = deepcopy(self.params)  # make sure it is idempotent

        # General settings can be be rendered without any problems
        section = params.pop("GeneralSettings")
        self._render_section(output, section)

        # System section has to make sure to keep the order persistent, this is why we use
        # self.system_order to iterate through
        section = params.pop("System")
        for my_id, name in enumerate(self.system_order):
            system = section[name]
            s_type = system.pop("type")
            output.append(f"{s_type} {my_id}")
            if s_type == "Framework":
                output.append(f"{' ' * 3}FrameworkName {name}")
            self._render_section(output, system, indent=3)

        # Section Component may contain parameters thar are for several systems
        # like:
        # CreateNumberOfMolecules n1, n2  <--- n1 is for system 0, n2 is for system 1
        section = params.pop("Component")
        for my_id, (name, molecule) in enumerate(section.items()):
            output.append(f"Component {my_id} MoleculeName {name}")
            if "BlockPocketsFileName" in molecule:
                if isinstance(molecule["BlockPocketsFileName"], dict):
                    try:
                        bps = self._dict_to_ordered_list(molecule["BlockPocketsFileName"])
                    except KeyError as err:
                        raise KeyError(
                            f"You did not provide BlockPocketsFileName parameter for the system '{str(err)}'"
                        ) from err

                    molecule["BlockPocketsFileName"] = bps
                    molecule["BlockPockets"] = ["yes" if bp else "no" for bp in bps]
                elif isinstance(molecule["BlockPocketsFileName"], str):
                    molecule["BlockPockets"] = "yes"

            if "CreateNumberOfMolecules" in molecule:
                if isinstance(molecule["CreateNumberOfMolecules"], dict):
                    try:
                        molecule["CreateNumberOfMolecules"] = self._dict_to_ordered_list(
                            molecule["CreateNumberOfMolecules"]
                        )
                    except KeyError as err:
                        raise KeyError(
                            f"You did not provide CreateNumberOfMolecules parameter for the system {str(err)}"
                        ) from err

            for item in ORDERED_ITEMS_COMPONENT_SECTION:
                if item in molecule:
                    self._render_item(output, item, molecule.pop(item), indent=3)

            self._render_section(output, molecule, indent=3)

        return "\n".join(output) + "\n"

    def _dict_to_ordered_list(self, input_dict):
        """Convert dict to ordered list.

        Convert dictionary {"framework_name1": something1, "framework_name2": something2}
        to [something2, something1], where the order is defined by self.system_order list

        :param input_dict: dictionary that looks as folows {"framework_name1": something1,
            "framework_name2": something2}
        :type input_dict: dict


        :returns: a list of items ordered according to the order of frameworks in the input file
        :rtype: list
        """
        ordered_list = [None] * len(self.system_order)
        for i, name in enumerate(self.system_order):
            ordered_list[i] = input_dict[name]
        return ordered_list

    # --------------------------------------------------------------------------
    @staticmethod
    def _render_section(output, params, indent=0):
        """
        It takes a dictionary and recurses through.

        For key-value pair it checks whether the value is a dictionary
        and prepends the key with &
        It passes the valued to the same function, increasing the indentation
        If the value is a list, I assume that this is something the user
        wants to store repetitively
        """
        #        output.append("enter")
        #        output.append("This what comes:" + str(params))
        for key, val in sorted(params.items()):
            RaspaInput._render_item(output, key, val, indent=indent)

    @staticmethod
    def _render_item(output, key, val, indent=0):
        """
        It takes one key-value item and adds to the output file
        """
        if isinstance(val, list):
            output.append(f"{' ' * indent}{key} {' '.join(str(p) for p in val)}")
        elif isinstance(val, bool):
            val_str = "yes" if val else "no"
            output.append(f"{' ' * indent}{key} {val_str}")
        else:
            output.append(f"{' ' * indent}{key} {val}")
